import sys
import os
import re

thisdir = os.path.dirname(os.path.abspath(__file__))
moddir = os.path.abspath(f"{thisdir}/../ext/c4core/cmake/bm-xp")
sys.path.insert(0, moddir)
import bm_plot as bm
from bm_util import first

from dataclasses import dataclass
import prettytable


def get_function_benchmark(libname, run: bm.BenchmarkRun):
    for rbm in run.entries:
        if rbm.meta.library == libname:
            return rbm
    raise Exception(f"lib not found: {libname}. Existing: {[rbm.meta.function for rbm in run.entries]}")


@dataclass
class ParseMeta:
    title: str
    library: str
    variant: str

    @classmethod
    def make(cls, bm_title: str):
        # eg:
        #-----------------------
        # RESERVE
        #   bm_ryml_inplace_reuse
        #   bm_ryml_arena_reuse
        #   bm_ryml_inplace
        #   bm_ryml_arena
        #   bm_libyaml_arena
        #   bm_libyaml_arena_reuse
        #   bm_libfyaml_arena
        #   bm_yamlcpp_arena
        #   bm_rapidjson_arena
        #   bm_rapidjson_inplace
        #   bm_sajson_arena
        #   bm_sajson_inplace
        #   bm_jsoncpp_arena
        #   bm_nlohmann_arena
        #-----------------------
        # EMIT
        #   bm_ryml_str_reserve
        #   bm_ryml_str
        #   bm_ryml_ostream
        #   bm_fyaml_str_reserve
        #   bm_fyaml_str
        #   bm_fyaml_ostream
        #   bm_yamlcpp
        #   bm_rapidjson
        #   bm_jsoncpp
        #   bm_nlohmann
        if not hasattr(__class__, '_rx'):
            __class__._rx = re.compile(r'bm_(ryml|libyaml|libfyaml|fyaml|yamlcpp|rapidjson|sajson|jsoncpp|nlohmann)_?(.*)')
        rx = __class__._rx
        if not rx.fullmatch(bm_title):
            raise Exception(f"cannot understand bm title: {bm_title}")
        lib = rx.sub(r'\1', bm_title)
        variant = rx.sub(r'\2', bm_title)
        return cls(
            title=bm_title,
            library=lib,
            variant=variant,
        )

    @property
    def shortname(self):
        return f"{self.library}_{self.variant}"

    @property
    def shortparams(self):
        return "params"

    @property
    def shorttitle(self):
        return self.shortname

EmitMeta = ParseMeta



def plot_bm_bars(bm_panel: bm.BenchmarkPanel, getref,
                 panel_title_human: str,
                 outputfile_prefix: str):
    assert os.path.isabs(outputfile_prefix), outputfile_prefix
    # make a comparison table
    atitle = lambda run: first(run.meta).shorttitle
    anchor = lambda run: f"{os.path.basename(outputfile_prefix)}-{atitle(run)}"
    anchorlink = lambda run: f"<pre><a href='#{anchor(run)}'>{atitle(run)}</a></pre>"
    with open(f"{outputfile_prefix}.txt", "w") as tablefile:
        with open(f"{outputfile_prefix}.md", "w") as mdfile:
            print(f"## {panel_title_human}\n\n<p>Data type benchmark results:</p>\n<ul>\n",
                  "\n".join([f"  <li>{anchorlink(run)}</li>" for run in bm_panel.runs]),
                  "</ul>\n\n", file=mdfile)
            for run in bm_panel.runs:
                tabletitle = f"{outputfile_prefix}"
                table = prettytable.PrettyTable(title=f"{panel_title_human}")
                table.add_column("function", [m.shorttitle for m in run.meta], align="l")
                for prop in ("mega_bytes_per_second", "cpu_time_ms"):
                    ref = getref(run)
                    bar_values = list(run.extract_plot_series(prop))
                    bar_values_rel = list(run.extract_plot_series(prop, relative_to_entry=ref))
                    bar_values_pc = list(run.extract_plot_series(prop, percent_of_entry=ref))
                    pd = bm_panel.first_run.property_plot_data(prop)
                    hns = pd.human_name_short
                    table.add_column(hns, [f"{v_:7.2f}" for v_ in bar_values], align="r")
                    hns = hns.replace(" (ms)", "")
                    table.add_column(f"{hns}(x)", [f"{v_:5.2f}x" for v_ in bar_values_rel], align="r")
                    table.add_column(f"{hns}(%)", [f"{v_:7.2f}%" for v_ in bar_values_pc], align="r")
                print(table, "\n\n")
                print(table, "\n\n", file=tablefile)
                pfx_bps = f"{os.path.basename(outputfile_prefix)}-mega_bytes_per_second"
                pfx_cpu = f"{os.path.basename(outputfile_prefix)}-cpu_time_ms"
                print(f"""
<br/>
<br/>

---

<a id="{anchor(run)}"/>

### {panel_title_human}

* Interactive html graphs
  * [MB/s](./{pfx_bps}.html)
  * [CPU time](./{pfx_cpu}.html)

[![{outputfile_prefix}: MB/s](./{pfx_bps}.png)](./{pfx_bps}.png)
[![{outputfile_prefix}: CPU time](./{pfx_cpu}.png)](./{pfx_cpu}.png)

```
{table}
```
""", file=mdfile)
    # make plots
    for prop in ("mega_bytes_per_second", "cpu_time_ms"):
        ps, ps_ = [], []
        pd = bm_panel.first_run.property_plot_data(prop)
        bar_label = f"{pd.human_name_short}{pd.qty_type.comment}"
        outfilename = f"{outputfile_prefix}-{prop}"
        for run in bm_panel.runs:
            bar_names = [m.shorttitle for m in run.meta]
            bar_values = list(run.extract_plot_series(prop))
            runtitle = f"{outfilename}"
            # to save each bokeh plot separately and also
            # a grid plot with all of them, we have to plot
            # twice because bokeh does not allow saving twice
            # the same plot from multiple pictures.
            plotit = lambda: bm.plot_benchmark_run_as_bars(run, title=f"{panel_title_human}\n{bar_label}",
                                                           bar_names=bar_names, bar_values=bar_values, bar_label=bar_label)
            # make one plot to save:
            p, p_ = plotit()
            bm._bokeh_save_html(f"{runtitle}.html", p)
            bm._plt_save_png(f"{runtitle}.png")
            bm._plt_clear()
            # and another to gather:
            p, p_ = plotit()
            ps.append(p)
            ps_.append(p_)
            bm._plt_clear()
        bm.bokeh_plot_many(ps, f"{outfilename}.html")




def plot_parse(dir_: str, filename, json_files):
    fcase = f"parse-{filename}"
    panel = bm.BenchmarkPanel(json_files, ParseMeta)
    ref = lambda bmrun: get_function_benchmark("yamlcpp", run=bmrun)
    plot_bm_bars(panel, ref,
                 f"parse benchmark: {filename}",
                 f"{dir_}/ryml-bm-{fcase}")


def plot_emit(dir_: str, filename, json_files):
    fcase = f"emit-{filename}"
    panel = bm.BenchmarkPanel(json_files, EmitMeta)
    ref = lambda bmrun: get_function_benchmark("yamlcpp", run=bmrun)
    plot_bm_bars(panel, ref,
                 f"emit benchmark: {filename}",
                 f"{dir_}/ryml-bm-{fcase}")


if __name__ == '__main__':
    args = sys.argv[1:]
    if len(args) < 2:
        raise Exception(f"usage: {sys.executable} {sys.argv[0]} <filename> <jsonfile>")
    cmd = args[0]
    print(cmd)
    filename = args[1]
    print(filename)
    json_files = args[2:]
    print(json_files)
    dir_ = os.path.dirname(json_files[0])
    for jf in json_files:
        print("jf:", jf, flush=True)
        assert os.path.dirname(jf) == dir_, (os.path.dirname(jf), dir_)
        assert os.path.exists(jf), jf
    if cmd == "emit":
        plot_emit(dir_, filename, json_files)
    elif cmd == "parse":
        plot_parse(dir_, filename, json_files)
    else:
        raise Exception(f"not implemented: {cmd}")



import re
import os
from os.path import abspath, dirname
import sys
import subprocess
import argparse
from typing import List
from enum import Enum

projdir = abspath(dirname(dirname(__file__)))
sys.path.insert(0, f"{projdir}/ext/c4core/cmake")
import amalgamate_utils as am
sys.path.insert(0, f"{projdir}/ext/c4core/tools")
import amalgamate as am_c4core


class Event(Enum):
    tree = "tree"
    testsuite = "testsuite"
    ints = "ints"
    ints_utils = "ints_utils"
    ints_to_testsuite = "ints_to_testsuite"
    all = "all"
    none = "none"
    def __str__(self):
        return self.value


event_doc = {
    Event.tree: """(the default) enable the normal ryml event handler
    to create the tree, and additionally the Tree, Node, parser and
    emitter utilities; if this is not enabled, none of these
    components will be included in the amalgamated file""",
    Event.testsuite: "enable the (extra) YAML test suite event handler",
    Event.ints: "enable the (extra) integer-based event handler",
    Event.ints_utils: "enable the (extra) integer-based event handler utils",
    Event.ints_to_testsuite: "enable the (extra) integer events conversion to testsuite events",
    Event.all: "enable all event handlers",
    Event.none: "disable all event handlers",
}


ryml_defmacro = "RYML_SINGLE_HDR_DEFINE_NOW"
c4core_defmacro = "C4CORE_SINGLE_HDR_DEFINE_NOW"
exports_def_code = f""" // shared library: export when defining
#if defined(RYML_SHARED) && defined({ryml_defmacro}) && !defined(RYML_EXPORTS)
#define RYML_EXPORTS
#endif
"""
c4core_def_code = f""" // propagate defines to c4core
#if defined({ryml_defmacro}) && !defined({c4core_defmacro})
#define {c4core_defmacro}
#endif

#if defined(RYML_EXPORTS) && !defined(C4CORE_EXPORTS)
#define C4CORE_EXPORTS
#endif

#if defined(RYML_SHARED) && !defined(C4CORE_SHARED)
#define C4CORE_SHARED
#endif

// workaround for include removal while amalgamating
// resulting in <stdarg.h> missing in arm-none-eabi-g++
// https://github.com/biojppm/rapidyaml/issues/193
#include <stdarg.h>
"""


def amalgamate_ryml(filename: str,
                    with_c4core: bool,
                    with_fastfloat: bool,
                    with_stl: bool,
                    events: List[Event]):
    c4core_amalgamated = ""
    if with_c4core:
        c4core_amalgamated = "src/c4/c4core_all.hpp"
        am_c4core.amalgamate_c4core(f"{projdir}/{c4core_amalgamated}",
                                    with_fastfloat=with_fastfloat,
                                    with_stl=with_stl)
    repo = "https://github.com/biojppm/rapidyaml"
    ryml_preamble = f"""
Rapid YAML - a library to parse and emit YAML, and do it fast.

{repo}

DO NOT EDIT. This file is generated automatically.
This is an amalgamated single-header version of the library.

INSTRUCTIONS:

  - Include at will in any header of your project. Because the
    amalgamated header file is large, to speed up compilation of
    your project, protect the include with its include guard
    `_RYML_SINGLE_HEADER_AMALGAMATED_HPP_`, ie like this:
    ```
    #ifndef _RYML_SINGLE_HEADER_AMALGAMATED_HPP_
    #include <header_file_name.hpp>
    #endif
    ```

  - In one -- and only one -- of your project source files, #define
    {ryml_defmacro} and then include this header. This will enable
    the function and class definitions in the header file.

  - To compile into a shared library, define the preprocessor symbol
    RYML_SHARED before including the header. This will take care of
    symbol export/import.

"""
    def has_evt(*which):
        if Event.all in events:
            return True
        for e in which:
            if e in events:
                return True
        return False
    srcfiles = [
        am.cmttext(ryml_preamble),
        am.cmtfile("LICENSE.txt"),
        am.injcode(exports_def_code),
        am.onlyif(with_c4core, am.injcode(c4core_def_code)),
        am.onlyif(with_c4core, c4core_amalgamated),
        "src/c4/yml/export.hpp",
        "src/c4/yml/fwd.hpp",
        "src/c4/yml/version.hpp",
        "src/c4/yml/common.hpp",
        "src/c4/yml/error.hpp",
        "src/c4/yml/error.def.hpp",
        "src/c4/yml/escape_scalar.hpp",
        "src/c4/yml/node_type.hpp",
        "src/c4/yml/tag.hpp",
        am.onlyif(has_evt(Event.tree), "src/c4/yml/tree.hpp"),
        am.onlyif(has_evt(Event.tree), "src/c4/yml/node.hpp"),
        am.onlyif(has_evt(Event.tree), "src/c4/yml/writer.hpp"),
        "src/c4/yml/detail/dbgprint.hpp",
        am.onlyif(has_evt(Event.tree), am.injcode("#define C4_YML_EMIT_DEF_HPP_")),
        am.onlyif(has_evt(Event.tree), "src/c4/yml/emit.hpp"),
        am.onlyif(has_evt(Event.tree), "src/c4/yml/emit.def.hpp"),
        "src/c4/yml/detail/stack.hpp",
        "src/c4/yml/filter_processor.hpp",
        "src/c4/yml/parser_state.hpp",
        "src/c4/yml/event_handler_stack.hpp",
        am.onlyif(has_evt(Event.tree), "src/c4/yml/event_handler_tree.hpp"),
        am.onlyif(has_evt(Event.ints), "src_extra/c4/yml/extra/event_handler_ints.hpp"),
        am.onlyif(has_evt(Event.ints_utils), "src_extra/c4/yml/extra/ints_utils.hpp"),
        am.onlyif(has_evt(Event.ints_to_testsuite), "src_extra/c4/yml/extra/ints_to_testsuite.hpp"),
        am.onlyif(has_evt(Event.testsuite), "src_extra/c4/yml/extra/string.hpp"),
        am.onlyif(has_evt(Event.testsuite), "src_extra/c4/yml/extra/event_handler_testsuite.hpp"),
        "src/c4/yml/parse_engine.hpp",
        "src/c4/yml/preprocess.hpp",
        am.onlyif(has_evt(Event.tree), "src/c4/yml/reference_resolver.hpp"),
        am.onlyif(has_evt(Event.tree), "src/c4/yml/parse.hpp"),
        am.onlyif(with_stl and has_evt(Event.tree), "src/c4/yml/std/map.hpp"),
        am.onlyif(with_stl and has_evt(Event.tree), "src/c4/yml/std/string.hpp"),
        am.onlyif(with_stl and has_evt(Event.tree), "src/c4/yml/std/vector.hpp"),
        am.onlyif(with_stl and has_evt(Event.tree), "src/c4/yml/std/std.hpp"),
        "src/c4/yml/version.cpp",
        "src/c4/yml/common.cpp",
        "src/c4/yml/node_type.cpp",
        "src/c4/yml/tag.cpp",
        "src/c4/yml/parse_engine.def.hpp",
        am.onlyif(has_evt(Event.tree), "src/c4/yml/tree.cpp"),
        am.onlyif(has_evt(Event.ints), "src_extra/c4/yml/extra/event_handler_ints.cpp"),
        am.onlyif(has_evt(Event.ints_utils), "src_extra/c4/yml/extra/ints_utils.cpp"),
        am.onlyif(has_evt(Event.ints_to_testsuite), "src_extra/c4/yml/extra/ints_to_testsuite.cpp"),
        am.onlyif(has_evt(Event.tree), "src/c4/yml/reference_resolver.cpp"),
        am.onlyif(has_evt(Event.tree), "src/c4/yml/parse.cpp"),
        am.onlyif(has_evt(Event.tree), "src/c4/yml/node.cpp"),
        "src/c4/yml/preprocess.cpp",
        am.onlyif(has_evt(Event.tree), "src/c4/yml/detail/checks.hpp"),
        am.onlyif(has_evt(Event.tree), "src/c4/yml/detail/print.hpp"),
        am.onlyif(has_evt(Event.tree), "src/c4/yml/yml.hpp"),
        am.onlyif(has_evt(Event.tree), "src/ryml.hpp"),
    ]
    result = am.catfiles(srcfiles,
                         projdir,
                         # comment out lines with these patterns:
                         include_regexes=[
                             re.compile(r'^\s*#\s*include "(c4/yml/.*)".*$'),
                             re.compile(r'^\s*#\s*include <(c4/yml/.*)>.*$'),
                             re.compile(r'^\s*#\s*include "(c4/.*)".*$'),
                             re.compile(r'^\s*#\s*include <(c4/.*)>.*$'),
                         ],
                         definition_macro=ryml_defmacro,
                         repo=repo,
                         result_incguard="_RYML_SINGLE_HEADER_AMALGAMATED_HPP_")
    result_with_only_first_includes = am.include_only_first(result)
    am.file_put_contents(filename, result_with_only_first_includes)


def mkparser():
    p = am.mkparser(
        c4core=(True, "amalgamate c4core together with ryml"),
        fastfloat=(True, "enable fastfloat library"),
        stl=(True, "enable stl interop")
    )
    default = [str(Event.tree)]
    evtdoc = '. '.join([f"'{e}': {event_doc[e]}" for e in Event])
    defaultdoc = ','.join([str(e) for e in default])
    p.add_argument('-e', '--events', type=str, default=default, choices=[str(e) for e in Event], nargs="+",
                   help=f"""Specify which event handlers to include. Possible
                   values are: {evtdoc}. The default is {defaultdoc}.""")
    return p


if __name__ == "__main__":
    args = mkparser().parse_args()
    args.events = [Event(e) for e in args.events] # is there a better way to do this?
    amalgamate_ryml(filename=args.output,
                    with_c4core=args.c4core,
                    with_fastfloat=args.fastfloat,
                    with_stl=args.stl,
                    events=args.events)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import os

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'rapidyaml'
copyright = '2018-2024 Joao Paulo Magalhaes <dev@jpmag.me>'
author = 'Joao Paulo Magalhaes <dev@jpmag.me>'
release = '0.11.1'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'doxygen', 'Thumbs.db', '.DS_Store']

# Tell sphinx what the primary language being documented is.
primary_domain = 'cpp'

# Tell sphinx what the pygments highlight language should be.
highlight_language = 'cpp'


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

#html_theme = 'alabaster'
html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

exclude_patterns += ["doxy_*.md"]

# ----------------------------------------------------------------------------
# Overwrite with the Doxygen documentation
# https://stackoverflow.com/a/41199722

"""
exclude_patterns += ["main.md", "include/*rst"]
import subprocess
this_dir = os.path.dirname(__file__)
subprocess.check_call(['doxygen', 'Doxyfile'], cwd=this_dir)
html_extra_path = ['./doxygen/html']
"""


# ----------------------------------------------------------------------------
# Setup the breathe extension
# https://breathe.readthedocs.io/en/latest/

# disabled as it was not working well with the groups
"""
extensions.append("breathe")
this_dir = os.path.dirname(__file__)
xml_dir = os.path.join(this_dir, "doxygen/xml")
assert os.path.exists(xml_dir)
breathe_projects = {
    "rapidyaml": xml_dir
}
breathe_default_project = "rapidyaml"
"""


# ----------------------------------------------------------------------------
# Setup the doxyrest extension
# https://vovkos.github.io/doxyrest/manual/basic.html
# https://github.com/vovkos/doxyrest

'''
# To get the binary:
# git clone --recursive https://github.com/vovkos/doxyrest_b
# cd doxyrest_b
# cmany i

# put the doxyrest dir in the path
import shutil
import sys
doxyrest_exe = shutil.which('doxyrest')
doxyrest_dir = os.path.abspath(os.path.dirname(os.path.dirname(doxyrest_exe)))
doxyrest_sphinx_dir = os.path.join(doxyrest_dir, "share/doxyrest/sphinx")
sys.path.insert(1, doxyrest_sphinx_dir)

extensions += ['doxyrest', 'cpplexer']

# commands to run
"""
set -xe ; \
cd ~/proj/rapidyaml ; \
cmany c -t Debug -c clang++ -V RYML_DEV=OFF -V RYML_BUILD_TESTS=ON ; \
cp -fav build/linux-x86_64-clangxx16.0-Debug/compile_commands.json doc ; \
cd doc ; \
make doxy ; \
export PATH=/home/jpmag/proj/doxyrest_b/install/linux-x86_64-gxx13.2-Release/bin:$PATH ; \
doxyrest -h ; \
doxyrest -c doxyrest-config.lua ; \
make html
"""
'''


# ----------------------------------------------------------------------------
# Setup the exhale extension
# https://exhale.readthedocs.io/en/latest/quickstart.html

# disabled as it was just not working will
"""
#extensions.append("exhale")
exhale_args = {
    # These arguments are required
    "containmentFolder":     "./api",
    "rootFileName":          "library_root.rst",
    "doxygenStripFromPath":  "..",
    # Heavily encouraged optional argument (see docs)
    "rootFileTitle":         "Library API",
    # Suggested optional arguments
    "createTreeView":        True,
    # TIP: if using the sphinx-bootstrap-theme, you need
    # "treeViewIsBootstrap": True,
    "exhaleExecutesDoxygen": True,
    "exhaleDoxygenStdin":    "INPUT = ../src"
}
"""
